import { Worker } from 'worker_threads'
import { v4 as uuidv4 } from 'uuid'
import os from 'os'

import logger from './logger'
// import { QuestionDb } from '../types/basicTypes'

// ---------------------------------------------------------------------------

interface WorkerObj {
  worker: any
  index: number
  free: Boolean
}

const jobCheckInterval = 500
const alertOnPendingCount = 10
const workerFile = './src/utils/classes.ts'
let workers: Array<WorkerObj>
const pendingJobs = {}
const completedJobs = {}

// ---------------------------------------------------------------------------

function handleWorkerError(worker: WorkerObj, err) {
  // TODO
  logger.Log('resourcePromise error', logger.GetColor('redbg'))
  console.error(err)
}

// TODO: accuire all workers here, and handle errors so they can be removed if threads exit
export function msgAllWorker(data: any): Promise<any> {
  console.log('MSGING ALL WORKER')
  return new Promise((resolve) => {
    const promises = []
    workers.forEach((worker) => {
      worker.worker.postMessage(data)
      console.log('MSGD')
      promises.push(
        new Promise((resolve) => {
          worker.worker.once('message', (msg) => {
            console.log(worker.index, 'ONCE MESSASGE RESOLVE')
            resolve(msg)
          })
        })
      )
    })
    Promise.all(promises).then((res) => {
      console.log('MSG ALL DONE', res)
      resolve(res)
    })
  })
}

export function doALongTask(obj: any): Promise<any> {
  if (Object.keys(pendingJobs).length > alertOnPendingCount) {
    logger.Log(
      `More than ${alertOnPendingCount} callers waiting for free resource! (${
        Object.keys(pendingJobs).length
      })`,
      logger.GetColor('redbg')
    )
  }

  const jobId = uuidv4()
  pendingJobs[jobId] = obj
  return new Promise((resolve) => {
    const interval = setInterval(() => {
      if (completedJobs[jobId] && completedJobs[jobId].done) {
        resolve(completedJobs[jobId].result)
        delete completedJobs[jobId]
        clearInterval(interval)
      }
    }, jobCheckInterval)
  })
}

export function initWorkerPool(initData: any): void {
  if (workers) {
    logger.Log('WORKERS ALREADY EXISTS', logger.GetColor('redbg'))
    return
  }
  workers = []
  const factory = {
    create: function(index) {
      return getAWorker(index, initData)
    },
    destroy: function(client) {
      // console.log('[DESTROY]')
      client.worker.terminate()
      // console.log('[DESTROYED] #' + client.index)
    },
  }

  const workerCount = process.env.NS_THREAD_COUNT || os.cpus().length - 1
  if (process.env.NS_WORKER_COUNT) {
    logger.Log(
      `Setting thread count from enviroment variable NS_WORKER_COUNT: '${workerCount}'`,
      logger.GetColor('red')
    )
  }

  for (let i = 0; i < workerCount; i++) {
    workers.push({
      worker: factory.create(i),
      index: i,
      free: true,
    })
  }

  // TODO: stop this interval sometime
  setInterval(() => {
    if (Object.keys(pendingJobs).length > 0) {
      const freeWorker = workers.find((worker) => {
        return worker.free
      })
      if (!freeWorker) {
        return
      }

      if (freeWorker.free) {
        freeWorker.free = false
      }
      // FIXME: FIFO OR ANYTHING ELSE (JOB PROCESSING ORDER)
      const workKey = Object.keys(pendingJobs)[0]
      const work = pendingJobs[workKey]
      completedJobs[workKey] = { done: false }
      delete pendingJobs[workKey]

      work.inProgress = true
      processJob(work, freeWorker).then((res) => {
        completedJobs[workKey].done = true
        completedJobs[workKey].result = res
      })
    }
  }, jobCheckInterval)
}

// ---------------------------------------------------------------------------

function processJob(work, worker) {
  return new Promise((resolve) => {
    doSomething(worker, work)
      .then((res) => {
        setTimeout(() => {
          resolve(res)
          // console.log('[RELEASE]: #' + client.index)
          worker.free = true
        }, 6000)
      })
      .catch(function(err) {
        handleWorkerError(worker, err)
      })
  })
}

function getAWorker(i, initData) {
  const worker = workerTs(workerFile, {
    workerData: {
      workerIndex: i,
      initData: initData,
    },
  })

  worker.setMaxListeners(50)

  worker.on('error', (err) => {
    logger.Log('Worker error!', logger.GetColor('redbg'))
    console.error(err)
  })

  worker.on('exit', (code) => {
    // TODO: this is critical, whole server should stop, or child threads should be restarted
    logger.Log(
      `[MAIN]: worker #${i} exit code: ${code}`,
      code === 0 ? logger.GetColor('redbg') : logger.GetColor('green')
    )
  })
  return worker
}

// ---------------------------------------------------------------------------

function doSomething(currWorker, obj) {
  const { /* index, */ worker } = currWorker
  return new Promise((resolve) => {
    // console.log('[ACCUIRE]: #' + index)
    worker.postMessage(obj)
    worker.once('message', (msg) => {
      resolve(msg)
    })
  })
}

const workerTs = (file: string, wkOpts: any) => {
  wkOpts.eval = true
  if (!wkOpts.workerData) {
    wkOpts.workerData = {}
  }
  wkOpts.workerData.__filename = file
  return new Worker(
    `
            const wk = require('worker_threads');
            require('ts-node').register();
            let file = wk.workerData.__filename;
            delete wk.workerData.__filename;
            require(file);
        `,
    wkOpts
  )
}
